/*-----------------------------------------------------------------------------
 
    >> INSTITUT INFORMACNICH TECHNOLOGII
	>> FAKULTA TECHNOLOGICKA
	>> UNIVERZITA TOMASE BATI VE ZLINE
					        
   	Soubor	:	CCurve.cpp
	Popis	:	Modelovani parametrickych krivek
	Autor	:	Jan Stalmach, stalmachjan@volny.cz 
	Obor	:	Informacni technologie, 2003-2004
  
    Implementovany nasledujci algoritmy:

	Bezierova krivka
	Bezierova racionalni krivka
	Coonsova BSpline krivka
	Catmul-Rom krivka
	Fergusonova krivka
	NURBS

	Bezierova kubika iterativne
	Bezierova kubika rekurzivne
	Coonsova kubika
	Catmul-Rom kubika
	Fergusonova kubika

	deCasteljau algoritmus

 ------------------------------------------------------------------------------*/


#include <math.h>
#include "CCurve.h"


/*---------------------------------------------------------------------
 konstruktor tridy CCurve
-----------------------------------------------------------------------*/
CCurve::CCurve(float r=1.0, float g=1.0, float b=1.0, int max_kroku=160) 
{ 
	Casovac = new Stopky;
	Cara	= new CLine(r,g,b);
	red   = r;
	green = g;
	blue  = b;
	MAX_KROKU = max_kroku;
}


/*---------------------------------------------------------------------
 nastavi vykreslovaci barvu krivky
-----------------------------------------------------------------------*/
void CCurve::Set_RGB(float r, float g, float b)
{
	red   = r;
	green = g;
	blue  = b;
	Cara->Set_RGB(red,green,blue);
}


/*---------------------------------------------------------------------
 Vykresli ridici polygon krivky zadane body p a stupnem polynomu n
-----------------------------------------------------------------------*/
void CCurve::Ridici_Polygon(bod *p, int stp, float rp, float gp, float bp)
{
	Cara->Set_RGB(rp,gp,bp);
	for(int i = 1;i<stp;i++) Cara->Bresenham(p[i-1].x, p[i-1].y,p[i].x, p[i].y);
}


/*---------------------------------------------------------------------
 Vypocet bodu Bezierovy kubiky
-----------------------------------------------------------------------*/
bod CCurve::Souradnice_Bezierova_Kubika_Iterativne(bod p1,bod p2,bod p3,bod p4, float t)
{
	float pom1, pom2, pom3;
	float B0,B1,B2,B3;
    bod point;

	// 4 bodova bezierova kubika je zadana nasledujici polynomialni rovnici:
	// B(t) = P1 * ( 1 - t )^3 + P2 * 3 * t * ( 1 - t )^2 + P3 * 3 * t^2 * ( 1 - t ) + P4 * t^3 

	// pomocne promene pro zprehledneni kodu
    pom1 = 1 - t;
    pom2 = pom1 * pom1 * pom1;
    pom3 = t * t * t;

	// Bersteinovy polynomy
	B0 = pom2;
	B1 = 3*t*pom1*pom1;
	B2 = 3*t*t*pom1;
	B3 = pom3;

	// vypocet bodu krivky pro t = <0,1>
    point.x = B0*p1.x + B1*p2.x + B2*p3.x + B3*p4.x;
    point.y = B0*p1.y + B1*p2.y + B2*p3.y + B3*p4.y;
    point.z = B0*p1.z + B1*p2.z + B2*p3.z + B3*p4.z;

	// vrat novou souradnici
    return point;				
}


/*---------------------------------------------------------------------
 Bezierova kubika urcena iterativnim postupem
-----------------------------------------------------------------------*/
double CCurve::Bezierova_Kubika_Iterativne(bod A, bod B, bod C, bod D)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	// pocatek krivky
	bod point0 = A;
	bod point1;

	for(float t = 0; t <= (1 + (1.0f / MAX_KROKU)); t += 1.0f / (MAX_KROKU - 1))
	{
		point1 = Souradnice_Bezierova_Kubika_Iterativne(A,B,C,D,t);
		Cara->Bresenham(point0.x, point0.y,point1.x, point1.y);
		point0 = point1;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*---------------------------------------------------------------------
 Algoritmus deCasteljau delici Bezierovu kubiku na dve casti
-----------------------------------------------------------------------*/
void CCurve::deCasteljau(bod *p, bod *left, bod *right)
{
	left[0]  = p[0];
	left[1]  = (p[0]+p[1])/2.0;
	bod hlp  = (p[1]+p[2])/2.0;
	left[2]	 = (left[1]+hlp)/2.0;
	right[3] = p[3];
	right[2] = (p[2]+p[3])/2.0;
	right[1] = (hlp+right[2])/2.0;
	left[3]  = right[0] = (left[2]+right[1])/2.0;
}


/*---------------------------------------------------------------------
 Urceni bodu Bezierovy kubiky pomoci rekursivniho algoritmu
-----------------------------------------------------------------------*/
void CCurve::Souradnice_Bezierova_Kubika_Rekursivne(bod *b, int level) 
{
	if (level<=0) Cara->Bresenham(b[0].x,b[0].y,b[3].x,b[3].y);
	else 
	{
		bod *left  = new bod[4];
		bod *right = new bod[4];

		deCasteljau(b,left,right);
				
		Souradnice_Bezierova_Kubika_Rekursivne(left, level-1);
		Souradnice_Bezierova_Kubika_Rekursivne(right, level-1);
	}
}
 

/*---------------------------------------------------------------------
 Bezierova kubika urcena rekursivnim postupem
-----------------------------------------------------------------------*/
double CCurve::Bezierova_Kubika_Rekursivne(bod A, bod B, bod C, bod D)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	bod *p = new bod[4];

	p[0]=A;
	p[1]=B;
	p[2]=C;
	p[3]=D;

	// Rekursivne generuje kubiku
	// MAX_KROKU je deleno 10, aby se eliminovala neprimerena rekurze
	Souradnice_Bezierova_Kubika_Rekursivne(p,(int)(MAX_KROKU/10)); 

	Casovac->Pause();
	return Casovac->GetTime();
}


/*---------------------------------------------------------------------
 Vypocet bodu Bezierovy krivky
-----------------------------------------------------------------------*/
bod CCurve::Souradnice_Bezier(bod *p,int n,float t)
{
  /* int kn,nn,nkn;
   float blend,muk,munk;
   
   bod b;

   b.x=0;
   b.y=0;
   b.z=0;

   muk = 1;
   munk = pow(1-t,(float)n);

   for (int k=0;k<=n;k++) 
   {
      nn = n;
      kn = k;
      nkn = n - k;
      blend = muk * munk;
      muk *= t;
      munk /= (1-t);
      while (nn >= 1)
	  {
         blend *= nn;
         nn--;
         if (kn > 1)
		 {
            blend /= (float)kn;
            kn--;
         }
         if (nkn > 1) 
		 {
            blend /= (float)nkn;
            nkn--;
         }
      }
      b.x += p[k].x * blend;
      b.y += p[k].y * blend;
      b.z += p[k].z * blend;
   }

   return(b);*/

   

   // jina moznost vypoctu, cca 3x pomalejsi nez vyse uvedeny algoritmus
   // tato metoda je stejna jako metoda pouzita pro vypocet racionalni krivky
   // pouze zde neuvazuji vahove koeficiety mi

   bod b = bod(0,0,0);
    
   for (int i=0; i<=n; i++)
   {
     b.x += p[i].x*Basis(n,i,t);
	 b.y += p[i].y*Basis(n,i,t);
	 b.z += p[i].z*Basis(n,i,t);
   }
   return b;

   
}


/*---------------------------------------------------------------------
 Bezierova krivka pro obecny polynom libovolneho stupne 
-----------------------------------------------------------------------*/
double CCurve::Bezier(bod *p,int stp)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	// pocatek krivky
	bod point0 = p[0];
	bod point1;

	for (float t = 0; t <= (1 + (1.0f / MAX_KROKU)); t += 1.0f / (MAX_KROKU - 1))
	{
		point1 = Souradnice_Bezier(p,stp-1,t);
		Cara->Bresenham(point0.x, point0.y,point1.x, point1.y);
		point0 = point1;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*-----------------------------------------------------------------------
 Fce pro vypocet faktorialu
-----------------------------------------------------------------------*/
float CCurve::Faktorial(int n)
{
   if (n==0) return 1; else return Faktorial(n-1)*n;
}


/*-----------------------------------------------------------------------
 Fce pro vypocet faktorialove funkce (kombinacnich cisel)
-----------------------------------------------------------------------*/
float CCurve::Ni(int n,int i)
{
    return Faktorial(n)/((Faktorial(i)*Faktorial(n-i)));
}


/*-----------------------------------------------------------------------
 Fce pro vypocet Bernsteinovych polynomu
-----------------------------------------------------------------------*/
float CCurve::Basis(int n,int i,float t)
{
    float ti;	//  t^i
    float tni;	//  (1 - t)^(n-i) 

    // vypocti t^i a (1-t)^(n-i)
    if ((t==0.)&&(i == 0)) ti=1.0; else  ti = pow(t,i);
    if ((n==i)&&(t==1.)) tni=1.0;  else tni = pow((1-t),(n-i));
	
    return Ni(n,i)*ti*tni;
}


/*---------------------------------------------------------------------
 Vypocet bodu Racionalni Bezierovy krivky
-----------------------------------------------------------------------*/
bod CCurve::Souradnice_Racionalni_Bezier(bod *p,int n,float t, float *m)
{
   float *R;
   R = new float[n];

   bod b = bod(0,0,0);

   float Value; 

   for (int i=0; i<=n; i++)
   {
     R[i]  = Basis(n,i,t)*m[i];  
	 Value = 0;
	 for (int j=0; j<=n; j++) Value += Basis(n,j,t)*m[j];
     b.x += R[i]*p[i].x/Value;
	 b.y += R[i]*p[i].y/Value;
	 b.z += R[i]*p[i].z/Value;
   }
   return b;
}


/*---------------------------------------------------------------------
 Racionalni Bezierova krivka pro obecny polynom libovolneho stupne 
-----------------------------------------------------------------------*/
double CCurve::Racionalni_Bezier(bod *p, int stp, float *m)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	// pocatek krivky
	bod point0 = p[0];
	bod point1;

	for (float t = 0; t <= (1 + (1.0f / MAX_KROKU)); t += 1.0f / (MAX_KROKU - 1))
	{
		point1 = Souradnice_Racionalni_Bezier(p,stp-1,t,m);
		Cara->Bresenham(point0.x, point0.y,point1.x, point1.y);
		point0 = point1;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*---------------------------------------------------------------------
 Vypocet bodu Fergusonovy kubiky
-----------------------------------------------------------------------*/
bod CCurve::Souradnice_Fergusonova_Kubika(bod p1,bod v1,bod v2,bod p2, float t)
{
	float pom1, pom2;
	float F1,F2,F3,F4;
    bod point;

	// pomocne promene pro zprehledneni kodu
	pom1 = t * t;
	pom2 = t * t * t;

	// kubicke Hermitovske polynomy
    F1 = 2*pom2 - 3*pom1 + 1;
    F2 = -2*pom2 + 3*pom1;
    F3 = pom2 - 2*pom1 + t;
    F4 = pom2 - pom1;

	// vypocet bodu krivky pro t = <0,1>
	point.x = F1*p1.x + F2*p2.x + F3*(v1.x - p1.x) + F4*(v2.x - p2.x);
    point.y = F1*p1.y + F2*p2.y + F3*(v1.y - p1.y) + F4*(v2.y - p2.y);
	point.z = F1*p1.z + F2*p2.z + F3*(v1.z - p1.z) + F4*(v2.z - p2.z);

	// vrat novou souradnici
    return point;				
}


/*---------------------------------------------------------------------
 Fergusonova kubika
-----------------------------------------------------------------------*/
double CCurve::Fergusonova_Kubika(bod A, bod vA, bod vB, bod B)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	// pocatek krivky
	bod point0 = A;
	bod point1;

	for(float t = 0; t <= (1 + (1.0f / MAX_KROKU)); t += 1.0f / (MAX_KROKU - 1))
	{
		point1 = Souradnice_Fergusonova_Kubika(A,vA,vB,B,t);
		Cara->Bresenham(point0.x, point0.y,point1.x, point1.y);
		point0 = point1;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*----------------------------------------------------------------------
 Fergusonova krivka tvorena napojovanim Fergusonovych kubickych oblouku
------------------------------------------------------------------------*/
double CCurve::Ferguson(bod *inp,int stp)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	int a=0;
	int b=1;
	int c=3;
	int d=2;

	for(int i = 0; i<stp-1; i+=2)
	{
		Fergusonova_Kubika(inp[a],inp[b],inp[c],inp[d]);	
		a=i;
		b=i+1;
		c=i+3;
		d=i+2;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*---------------------------------------------------------------------
 Vypocet bodu Coonsovy kubiky
-----------------------------------------------------------------------*/
bod CCurve::Souradnice_Coonsova_Kubika(bod p1,bod p2,bod p3,bod p4, float t)
{
	float pom1, pom2, pom3;
	float C1,C2,C3,C4;
    bod point;

	// Pomocne promene pro zprehledneni kodu
	pom1 = 1 - t;
	pom2 = t * t;
	pom3 = pom1 * pom1 * pom1;

	// Polynomy
    C1 = pom3 / 6;
    C2 = (3*pom2*t - 6*pom2 + 4) / 6;
    C3 = (-3*pom2*t + 3*pom2 +3*t +1) / 6;
    C4 = pom2*t / 6;

	// Vypocet bodu krivky pro t = <0,1>
	point.x = C1*p1.x + C2*p2.x + C3*p3.x + C4*p4.x;
    point.y = C1*p1.y + C2*p2.y + C3*p3.y + C4*p4.y;
	point.z = C1*p1.z + C2*p2.z + C3*p3.z + C4*p4.z;

	// vrat novou souradnici
    return point;				
}


/*---------------------------------------------------------------------
 Coonsova kubika
-----------------------------------------------------------------------*/
double CCurve::Coonsova_Kubika(bod A, bod B, bod C, bod D)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	// pocatek krivky nelezi v bode A
	bod point0((A.x+(4*B.x)+C.x)/6,(A.y+(4*B.y)+C.y)/6,(A.z+(4*B.z)+C.z)/6);
	bod point1;


	for(float t = 0; t <= (1 + (1.0f / MAX_KROKU)); t += 1.0f / (MAX_KROKU-1))
	{
		point1 = Souradnice_Coonsova_Kubika(A,B,C,D,t);
		Cara->Bresenham(point0.x, point0.y,point1.x, point1.y);
		point0 = point1;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*-------------------------------------------------------------------------
 Coonsova B-Spline krivka tvorena napojovanim Coonsovych kubickych oblouku
---------------------------------------------------------------------------*/
double CCurve::BSplineCoons(bod *inp,int stp)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	for(int i = 0; i<stp-3; i++)
	{
		Coonsova_Kubika(inp[i],inp[i+1],inp[i+2],inp[i+3]);
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*---------------------------------------------------------------------
 Vypocet bodu Catmul-Rom kubiky
-----------------------------------------------------------------------*/
bod CCurve::Souradnice_Catmul_Rom_Kubika(bod p1,bod p2,bod p3,bod p4, float t)
{
	float pom1, pom2;
	float C1,C2,C3,C4;
    bod point;

	// Pomocne promene pro zprehledneni kodu
	pom1 = t - 1;
	pom2 = t * t;

	// Polynomy
    C1 = (-pom2*t + 2*pom2 - t)  / 2;
    C2 = (3*pom2*t - 5*pom2 + 2) / 2;
    C3 = (-3*pom2*t + 4*pom2 +t) / 2;
    C4 = pom1*pom2 / 2;

	// Vypocet bodu krivky pro t = <0,1>
	point.x = C1*p1.x + C2*p2.x + C3*p3.x + C4*p4.x;
    point.y = C1*p1.y + C2*p2.y + C3*p3.y + C4*p4.y;
	point.z = C1*p1.z + C2*p2.z + C3*p3.z + C4*p4.z;

	// vrat novou souradnici
    return point;				
}


/*---------------------------------------------------------------------
 Catmul-Rom
-----------------------------------------------------------------------*/
double CCurve::Catmul_Rom_Kubika(bod A, bod B, bod C, bod D)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	// pocatek krivky lezi v bode B
	bod point0=B;
	bod point1;


	for(float t = 0; t <= (1 + (1.0f / MAX_KROKU)); t += 1.0f / (MAX_KROKU-1))
	{
		point1 = Souradnice_Catmul_Rom_Kubika(A,B,C,D,t);
		Cara->Bresenham(point0.x, point0.y,point1.x, point1.y);
		point0 = point1;
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*-------------------------------------------------------------------------
 Catmul-Rom krivka tvorena napojovanim Catmul-Rom kubickych oblouku
---------------------------------------------------------------------------*/
double CCurve::Catmul_Rom(bod *inp,int stp)
{
	Barva(red,green,blue);
	Casovac->Unpause();

	for(int i = 0; i<stp-3; i++)
	{
		Catmul_Rom_Kubika(inp[i],inp[i+1],inp[i+2],inp[i+3]);
	}

	Casovac->Pause();
	return Casovac->GetTime();
}


/*---------------------------------------------------------------------
   Rekurentni vypocet B-Spline bazove fce
-----------------------------------------------------------------------*/
double CCurve::BSpline_baze(int k,int t,float *u,double v)
{
	double value;

	// pro t=1 nastav primo hodnotu (podle definice B-Spline fce)
    if (t==1) 
    {
       if ((u[k] <= v) && (v < u[k+1])) value = 1; else value = 0;
    } 

    if (t>1) 
	{
		// pokud jsou obe casti rekurentni rce rovny nule tak vrat nulu
		if ((u[k+t-1] == u[k]) && (u[k+t] == u[k+1])) value = 0; 
		
		// prvni cast rovnice je nula
		else if (u[k+t-1] == u[k]) value = (u[k+t] - v) / (u[k+t] - u[k+1]) * BSpline_baze(k+1,t-1,u,v);
		
		// druha cast rovnice je nula
		else if (u[k+t] == u[k+1]) value = (v - u[k]) / (u[k+t-1] - u[k]) * BSpline_baze(k,t-1,u,v); 
		
		// zadna cast rovnice neni nula
		else
        value = ((v - u[k]) / (u[k+t-1] - u[k])) * BSpline_baze(k,t-1,u,v) + 
                ((u[k+t] - v) / (u[k+t] - u[k+1])) * BSpline_baze(k+1,t-1,u,v);
    }

    return(value);
}

/*---------------------------------------------------------------------
   This returns the point "output" on the spline curve.
   The parameter "v" indicates the position, it ranges from 0 to n-t+2   
-----------------------------------------------------------------------*/
void CCurve::BSpline_bod(float *u,int n,int t,double v,bod *control,bod *output)
{
   double b;

   output->x = 0;
   output->y = 0;
   output->z = 0;

   for (int k=0;k<=n;k++) 
   {
      b = BSpline_baze(k,t,u,v);
      output->x += control[k].x * b;
      output->y += control[k].y * b;
      output->z += control[k].z * b;
   }
}


/*---------------------------------------------------------------------
   The positions of the subintervals of v and breakpoints, the position
   on the curve are called knots. Breakpoints can be uniformly defined
   by setting u[j] = j, a more useful series of breakpoints are defined
   by the function below. This set of breakpoints localises changes to
   the vicinity of the control point being modified.

 Uniformni uzlovy vektor u pro B-Spline fci
-----------------------------------------------------------------------*/
void CCurve::BSpline_uzly(float *u,int n,int t)
{
   int j;

   for (j=0;j<=n+t;j++) 
   {
      if      (j < t)  u[j] = 0;
      else if (j <= n) u[j] = j - t + 1;
      else if (j > n)  u[j] = n - t + 2;
   }
}


/*---------------------------------------------------------------------
 Generovani bodu BSpline krivky
-----------------------------------------------------------------------*/
void CCurve::Souradnice_BSpline(bod *inp,int n,float *uzly,int t,bod *outp,int res)
{
   double interval,inc;

   interval = 0;
   inc = (n - t + 2) / (double)(res - 1);
   for (int i=0;i<res-1;i++) 
   {
      BSpline_bod(uzly,n,t,interval,inp,&(outp[i]));
      interval += inc;
   }
   outp[res-1] = inp[n];
}


/*---------------------------------------------------------------------
 B-Spline krivka pro obecny polynom libovolneho stupne 
-----------------------------------------------------------------------*/
double CCurve::BSpline(bod *inp,int n, int t)
{
    Barva(red,green,blue);
    Casovac->Unpause();

	if ((n<4)||(t>n)) return 0;

    uzel = new float[n+t];

    bod *outp;
	outp = new bod[MAX_KROKU];
   
	BSpline_uzly(uzel,n-1,t);

	//for (int i=0;i<n+t;i++) {cout << uzly[i] << " ";}
	//cout << endl;

    Souradnice_BSpline(inp,n-1,uzel,t,outp,MAX_KROKU);

	for (int i=1;i<MAX_KROKU;i++) 
	{
		Cara->Bresenham(outp[i-1].x,outp[i-1].y,outp[i].x,outp[i].y);
	}

   Casovac->Pause();
   return Casovac->GetTime();
}


/*---------------------------------------------------------------------
 Bod NURBS krivky
-----------------------------------------------------------------------*/
void CCurve::NURBS_bod(float *u,float *vaha, int n,int t,double v,bod *control,bod *output)
{
   double b;

   output->x = 0;
   output->y = 0;
   output->z = 0;

   bod jmen = bod(0,0,0);

   for (int k=0;k<=n;k++) 
   {
      b = BSpline_baze(k,t,u,v)*vaha[k];
      output->x += control[k].x * b;
      output->y += control[k].y * b;
      output->z += control[k].z * b;
	  jmen.x += b;
	  jmen.y += b;
      jmen.z += b;
   }
   output->x /= jmen.x;
   output->y /= jmen.y;
   output->z /= jmen.z;
}


/*---------------------------------------------------------------------
 Neuniformni uzlovy vektor u pro B-Spline NURBS fci
-----------------------------------------------------------------------*/
void CCurve::NURBS_uzly(int *u,int n,int t)
{
   int j;

   for (j=0;j<=n+t;j++) 
   {
      if      (j < t)  u[j] = 0;
      else if (j <= n) u[j] = j - t + 1;
      else if (j > n)  u[j] = n - t + 2;
   }
}


/*---------------------------------------------------------------------
 Generovani bodu NURBS krivky
-----------------------------------------------------------------------*/
void CCurve::Souradnice_NURBS(bod *inp,int n,float *uzly,float *vaha,int t,bod *outp,int res)
{
   double interval,inc;

   interval = 0;
   inc = 1 / (double)(res - 1);
   for (int i=0;i<res-1;i++) 
   {
      NURBS_bod(uzly,vaha,n,t,interval,inp,&(outp[i]));
      interval += inc;
   }
   outp[res-1] = inp[n];
}



/*---------------------------------------------------------------------
 NURBS
-----------------------------------------------------------------------*/
double CCurve::NURBS(bod *inp,float *uzly, float *vaha,int n, int U_size)
{
	Barva(red,green,blue);
    Casovac->Unpause();

	//if ((n<4)||(t>n)) return 0;

	//int *uzly;
    //uzly = new int[stp+t];

	bod *outp;
	outp = new bod[MAX_KROKU];
   
	//NURBS_uzly(uzly,stp-1,t);

    Souradnice_NURBS(inp,n-1,uzly,vaha,U_size-n,outp,MAX_KROKU);

	for (int i=1;i<MAX_KROKU;i++) 
	{
		Cara->Bresenham(outp[i-1].x,outp[i-1].y,outp[i].x,outp[i].y);
	}

   Casovac->Pause();
   return Casovac->GetTime();
}